# ML panel data models
# Linear model
# AR(1) - fixed effects and random effects
  # Serial autocorrelation model

# THESE PACKAGES MAY BE NEEDED
  # library(mvtnorm)
  # library(ramify)
  # library(tidyr)
  # library(dplyr)

# Simple linear model
LL_reg <- function(B, y, X) {
  N<-nrow(X)
  b <- B[1:(length(B)-1)]
  lsig <- B[length(B)]
  
  y_std <- (y - X %*% b)/exp(lsig/2)
  # LL enters with a minus
  LL <- sum(0.5*(lsig + log(2*pi) + y_std^2))
  return(LL)
}

# We assume balanced panel
LL_RE <- function(B, y, X, TT) {
  
  N <- nrow(X)/TT
  
  b <- B[1:(length(B)-2)]
  #sig1 <- exp(B[length(B)-1]) # RE
  #sig2 <- exp(B[length(B)]) # Error term
  sig1 <- exp(B[length(B)-1])
  rho <- exp(B[length(B)])/(exp(B[length(B)]) +1) # rho
  
  E <- (y - X %*% b)
  E <- t(matrix(E, nrow = TT, ncol = N)) # N x TT
  
  #COV<-diag(TT)*(sig1+sig2)
  # COV[COV == 0] <- sig1
  COV<-diag(TT)*(sig1)
  COV[COV == 0] <- sig1*rho
  # Tmp <-(E %*% solve(COV))
 # print(COV)
  Tmp <-solve(COV, t(E))
  Vec  <- matrix(, nrow = N, ncol = 1)
  for(i in 1:N){
    # Vec[i] <- Tmp[i,] %*% t(E)[,i]
    Vec[i] <- E[i,] %*% Tmp[,i]
  }
  ldet<-determinant(COV, logarithm = TRUE)
  
  LL <- sum(0.5*(TT*log(2*pi) + ldet$modulus + Vec))
  return(LL)
}

LL_REAR <- function(B, y, X, TT) {
  
  N <- nrow(X)/TT
  b <- B[1:(length(B)-3)]
  #sig1 <- exp(B[length(B)-1]) # RE
  #sig2 <- exp(B[length(B)]) # Error term
  
  sig1 <- exp(B[length(B)-2])
  rho <- exp(B[length(B)-1])/(exp(B[length(B)-1]) +1) # rho for RE
  rho2 <- 2*exp(B[length(B)])/(exp(B[length(B)]) +1) -1 # rho for the lag
  
  Fit <- X %*% b
  A <-diag(TT) 
  A[2:(TT),1:(TT-1)] <- A[2:(TT),1:(TT-1)]-rho2*diag(TT-1)
  Mat <- solve(A)
  
  Fit <- matrix(Fit, nrow = TT, ncol = N) # TT x N
  Means <- Mat %*% Fit # Means
  y <- matrix(y, nrow = TT, ncol = N) # TT x N
  E <- t(y - Means) # N x TT
  
  #COV<-diag(TT)*(sig1+sig2)
  # COV[COV == 0] <- sig1
  COV<-diag(TT)*(sig1)
  COV[COV == 0] <- sig1*rho
  #COV[1,1]<- COV[1,1] + (1-rho)*sig1*(rho2^2) # y0 is only eps
  COV[1,1]<- COV[1,1] + sig1*(rho2^2) + 2*rho2*sig1*rho # y0 is eps + re
  COV[2:TT,1]<-COV[2:TT,1] + sig1*rho*rho2
  COV[1, 2:TT]<-COV[1, 2:TT] + sig1*rho*rho2
  
  VAR<- Mat %*% COV %*% t(Mat)
  
  
  Tmp <-solve(VAR, t(E), tol=0)
  Vec  <- matrix(, nrow = N, ncol = 1)
  
  for(i in 1:N){
    # Vec[i] <- Tmp[i,] %*% t(E)[,i]
    Vec[i] <- E[i,] %*% Tmp[,i]
  }
  ldet<-determinant(VAR, logarithm = TRUE)
  
  LL <- sum(0.5*(TT*log(2*pi) + ldet$modulus + Vec))
  return(LL)
}

LL_FD <- function(B, y, X, TT) {
  N <- nrow(X)/TT
  b <- B[1:(length(B)-1)]
  sig1 <- exp(B[length(B)])
  
  E <- (y - X %*% b)
  E <- t(matrix(E, nrow = TT, ncol = N)) # N x TT
  
  COV <- 2*diag(TT)*(sig1)
  COV[2:(TT),1:(TT-1)] <- COV[2:(TT),1:(TT-1)]-sig1*diag(TT-1)
  COV[1:(TT-1), 2:(TT)] <- COV[1:(TT-1), 2:(TT)]-sig1*diag(TT-1)
  
  #  if (det(COV) <= 0){
  #    COV <- nearPD(COV)$mat
  #  COV <- make.positive.definite(COV)
  #  }
  
  # L<-chol(COV)
  Tmp <-solve(COV, t(E), tol=0)
  #  Tmp <-solve(COV, tol = 0) %*% t(E)
  
  Vec  <- matrix(, nrow = N, ncol = 1)
  
  for(i in 1:N){
    # Vec[i] <- crossprod(backsolve(t(L), E[i,]))
    Vec[i] <- E[i,] %*% Tmp[,i]
  }
  ldet<-determinant(COV, logarithm = TRUE)
  
  #ldet<- log(det(COV))
  LL <- sum(0.5*(TT*log(2*pi) + ldet$modulus + Vec))
  #  LL <- sum(0.5*(TT*log(2*pi) + ldet + Vec))
  return(LL)
}

LL_FDAR <- function(B, y, X, TT) {
  N <- nrow(X)/TT
  b <- B[1:(length(B)-2)]
  sig1 <- exp(B[length(B)-1])
  rho2 <- 2*exp(B[length(B)])/(exp(B[length(B)]) +1) -1 # rho for the lag
  
  Fit <- X %*% b
  A <-diag(TT) 
  A[2:(TT),1:(TT-1)] <- A[2:(TT),1:(TT-1)]-rho2*diag(TT-1)
  Mat <- solve(A)
  
  Fit <- matrix(Fit, nrow = TT, ncol = N) # TT x N
  Means <- Mat %*% Fit # Means
  
  y <- matrix(y, nrow = TT, ncol = N) # TT x N
  E <- t(y - Means) # N x TT
  COV <- 2*diag(TT)*(sig1)
  COV[2:(TT),1:(TT-1)] <- COV[2:(TT),1:(TT-1)]-sig1*diag(TT-1)
  COV[1:(TT-1), 2:(TT)] <- COV[1:(TT-1), 2:(TT)]-sig1*diag(TT-1)
  #COV[1,1] <- COV[1,1] + 2*sig1*rho2^2 - 2*sig1*rho2
  VAR<- Mat %*% COV %*% t(Mat)
  
  # L<-chol(VAR)
  # print(VAR)
  # print(t(L) %*% L)
 # VAR <- make.positive.definite(VAR)
  Tmp <-solve(VAR, t(E), tol=0)
  Vec  <- matrix(, nrow = N, ncol = 1)
  
  for(i in 1:N){
    #  Vec[i] <- crossprod( backsolve(t(L), E[i,]))
    Vec[i] <- E[i,] %*% Tmp[,i]
  }
  #ldet<-determinant(VAR, logarithm = TRUE)
  #LL <- sum(0.5*(TT*log(2*pi) + ldet$modulus + Vec))
  ldet<- log(det(COV))
    LL <- sum(0.5*(TT*log(2*pi) + ldet + Vec))
  
  return(LL)
}

LL_FDAR2 <- function(B, y, X, TT, X2) {
  N <- nrow(X)/TT
  b <- B[1:(length(B)-2)]
  sig1 <- exp(B[length(B)-1])
  rho2 <- 2*exp(B[length(B)])/(exp(B[length(B)]) +1) -1 # rho for the lag
  
  Fit <- X %*% b
  Fit2 <- rho2*(X2 %*% b)
  Fit <- Fit + Fit2
  A <-diag(TT) 
  A[2:(TT),1:(TT-1)] <- A[2:(TT),1:(TT-1)]-rho2*diag(TT-1)
  Mat <- solve(A)
  
  Fit <- matrix(Fit, nrow = TT, ncol = N) # TT x N
  Means <- Mat %*% Fit # Means
  
  y <- matrix(y, nrow = TT, ncol = N) # TT x N
  E <- t(y - Means) # N x TT
  COV <- 2*diag(TT)*(sig1)
  COV[2:(TT),1:(TT-1)] <- COV[2:(TT),1:(TT-1)]-sig1*diag(TT-1)
  COV[1:(TT-1), 2:(TT)] <- COV[1:(TT-1), 2:(TT)]-sig1*diag(TT-1)
 # COV[1,1] <- COV[1,1] + 2*sig1*rho2^2 - 2*sig1*rho2 # Current version
  COV[1,1] <- sig1 + sig1*(1-rho2)^2 + sig1*((1-rho2)^2)*rho2^2 + sig1*rho2^4
  
  # COV[1,1] <- COV[1,1] + sig1*(1+rho2^2)*(1-rho2)^2
  
  VAR<- Mat %*% COV %*% t(Mat)
  
  # L<-chol(VAR)
  # print(VAR)
  # print(t(L) %*% L)
  # VAR <- make.positive.definite(VAR)
  Tmp <-solve(VAR, t(E), tol=0)
  Vec  <- matrix(, nrow = N, ncol = 1)
  
  for(i in 1:N){
    #  Vec[i] <- crossprod( backsolve(t(L), E[i,]))
    Vec[i] <- E[i,] %*% Tmp[,i]
  }
  #ldet<-determinant(VAR, logarithm = TRUE)
  #LL <- sum(0.5*(TT*log(2*pi) + ldet$modulus + Vec))
  ldet<- log(det(COV))
  LL <- sum(0.5*(TT*log(2*pi) + ldet + Vec))
  
  return(LL)
}

LL_FDAR3 <- function(B, y, X, TT, yfr) {
  N <- nrow(X)/TT
  b <- B[1:(length(B)-2)]
  sig1 <- exp(B[length(B)-1])
  rho2 <- 2*exp(B[length(B)])/(exp(B[length(B)]) +1) -1 # rho for the lag
  
  Fit <- X %*% b
  A <-diag(TT) 
  A[2:(TT),1:(TT-1)] <- A[2:(TT),1:(TT-1)]-rho2*diag(TT-1)
  Mat <- solve(A)
  
  Fit <- matrix(Fit, nrow = TT, ncol = N) # TT x N
  Means <- Mat %*% Fit # Means
  
  y <- matrix(y, nrow = TT, ncol = N) # TT x N
  yfr <- matrix(yfr, nrow = TT, ncol = N) # TT x N
  y[1,]<- y[1,] - rho2*yfr[1,]
  E <- t(y - Means) # N x TT
  COV <- 2*diag(TT)*(sig1)
  COV[2:(TT),1:(TT-1)] <- COV[2:(TT),1:(TT-1)]-sig1*diag(TT-1)
  COV[1:(TT-1), 2:(TT)] <- COV[1:(TT-1), 2:(TT)]-sig1*diag(TT-1)
  COV[1,1] <- COV[1,1] + (sig1)*rho2^2
  VAR<- Mat %*% COV %*% t(Mat)
  
  # L<-chol(VAR)
  # print(VAR)
  # print(t(L) %*% L)
  # VAR <- make.positive.definite(VAR)
  Tmp <-solve(VAR, t(E), tol=0)
  Vec  <- matrix(, nrow = N, ncol = 1)
  
  for(i in 1:N){
    #  Vec[i] <- crossprod( backsolve(t(L), E[i,]))
    Vec[i] <- E[i,] %*% Tmp[,i]
  }
  #ldet<-determinant(VAR, logarithm = TRUE)
  #LL <- sum(0.5*(TT*log(2*pi) + ldet$modulus + Vec))
  ldet<- log(det(COV))
  LL <- sum(0.5*(TT*log(2*pi) + ldet + Vec))
  
  return(LL)
}

FirstDiff <- function(data){
  J<-ncol(data)
  for (i in 2:J){
    data<-data %>%
      group_by(id) %>%
      mutate(across(colnames(data)[i], ~dplyr::lag(.), .names = "{col}_l"))
    data[,colnames(data)[i]] <- data[,colnames(data)[i]] - data[,ncol(data)]
  }
  data<-data.frame(data[!is.na(data[,2]),])
  return(data[,1:J])
}

FirstRow <- function(data){
  J<-ncol(data)
  for (i in 2:J){
    data<-data %>%
      group_by(id) %>%
      mutate(across(colnames(data)[i], ~dplyr::lag(.), .names = "{col}_l"))
    
    data[!is.na(data[,ncol(data)]),colnames(data)[i]] <- 0
  }
  data<- data %>% group_by(id) %>% filter(row_number() != n())
  
  return(data.frame(data[,2:J]))
}
# Main function for estimating panel models
  # Types are: LM, RE, REAR, FD, FDAR
MLP <- function(formula, data, TT = 1, index, type = "LM", 
                start = matrix(, nrow = 1, ncol = 1)) {
  
  X <- model.matrix(formula, data)
  # y<-data[,as.character(formula[[2]])]
  y<-data.matrix(data[,as.character(formula[[2]])])
  
 
  if(type == "FDAR3" | type == "FDAR2" | type == "FDAR" | type == "FD"){
    id <- data[,index]
    names(id)<- "id"
    Xfr<- FirstRow(data.frame(id, y, X[,2:ncol(X)]))
    names(Xfr)[1]<-as.character(formula[[2]])
    yfr<- Xfr[,as.character(formula[[2]])]
    Xfr <- model.matrix(formula, Xfr)
    Xfr <-Xfr[,2:ncol(X)]
    
    data_diff <- FirstDiff(data.frame(id, y, X[,2:ncol(X)]))
    names(data_diff)[2]<-as.character(formula[[2]])
    X <- model.matrix(formula, data_diff)
    X<-X[,2:ncol(X)]
    y<-data.matrix(data_diff[,as.character(formula[[2]])])
    #y<-data_diff[,as.character(formula[[2]])]
    
    TT<-TT - 1
  }
  
  k <- ncol(X) # no. of coefficients
  
  if(type == "LM"){
    res<-optim(rep(0.01,k+1), LL_reg,method = "BFGS", hessian=T,y=y,X=X)
    coefnames <- c(colnames(X), "Sigma")
  } else if (type == "RE"){
    tmp<- optim(rep(0.01,k+1), LL_reg,method = "BFGS", y=y,X=X)
    if (all(is.na(start))){
      start <- c(tmp$par[1:k], tmp$par[k+1], -0.01)
    }
    res<-optim(start, LL_RE, hessian=T,y=y,X=X, TT = TT)
    # ,method="BFGS"
    coefnames <- c(colnames(X), "Sigma", "RhoRE")
   } else if (type == "REAR"){
      tmp<- optim(rep(0.01,k+1), LL_reg,method="BFGS", y=y,X=X)
      if (all(is.na(start))){
        start <- c(tmp$par[1:k], tmp$par[k+1], -0.01, 0.01)
      }
      res<-optim(start, LL_REAR, hessian=T,y=y,X=X, TT = TT)
      coefnames <- c(colnames(X), "Sigma", "RhoRE", "RhoLag")
   } else if (type == "FD"){
     tmp<- optim(rep(0.01,k+1), LL_reg,y=y,X=X)
     if (all(is.na(start))){
       start <- c(tmp$par[1:k], tmp$par[k+1])
     }
     res<-optim(start, LL_FD, hessian=T,y=y,X=X, TT = TT)
     coefnames <- c(colnames(X), "Sigma")
   } else if (type == "FDAR"){
     tmp<- optim(rep(0.01,k+1), LL_reg,y=y,X=X)
     if (all(is.na(start))){
       start <- c(tmp$par[1:k], tmp$par[k+1]-log(2), 0.1)
     }
     res<-optim(start, LL_FDAR, hessian=T,y=y, X=X, TT = TT)
     coefnames <- c(colnames(X), "Sigma", "RhoLag")
   } else if (type == "FDAR2"){

     tmp<- optim(rep(0.01,k+1), LL_reg,y=y,X=X)
     if (all(is.na(start))){
       start <- c(tmp$par[1:k], tmp$par[k+1]-log(2), 0.1)
     }
     res<-optim(start, LL_FDAR2, hessian=T,y=y, X=X, TT = TT, X2 = Xfr)
     coefnames <- c(colnames(X), "Sigma", "RhoLag")
   } else if (type == "FDAR3"){
     
     tmp<- optim(rep(0.01,k+1), LL_reg,y=y,X=X)
     if (all(is.na(start))){
       start <- c(tmp$par[1:k], tmp$par[k+1]-log(2), 0.1)
     }
     res<-optim(start, LL_FDAR3, hessian=T,y=y, X=X, TT = TT, yfr = yfr)
     coefnames <- c(colnames(X), "Sigma", "RhoLag")
  }
  
  b<-res$par
  VC<-solve(res$hessian)
  se<-sqrt(diag(VC))
  
  b[coefnames == "Sigma"] <- exp(b[coefnames == "Sigma"])
  se[coefnames == "Sigma"] <- b[coefnames == "Sigma"]*se[coefnames == "Sigma"]
  b[coefnames == "RhoRE"] <- exp(b[coefnames == "RhoRE"])/(1+exp(b[coefnames == "RhoRE"]))
  se[coefnames == "RhoRE"] <- se[coefnames == "RhoRE"]*exp(b[coefnames == "RhoRE"])/(1+exp(b[coefnames == "RhoRE"]))^2
  b[coefnames == "RhoLag"] <- 2*exp(b[coefnames == "RhoLag"])/(1+exp(b[coefnames == "RhoLag"])) -1
  se[coefnames == "RhoLag"] <- 2*se[coefnames == "RhoLag"]*exp(b[coefnames == "RhoLag"])/(1+exp(b[coefnames == "RhoLag"]))^2
  
  z<-b/se
  pval<-2*(1-pnorm(abs(z)))
  results<-cbind(b,se,z,pval)
  colnames(results)<-c("b","se","z","p")
  rownames(results)<- coefnames
  print(results,digits=3)
  return(res)
}

# Function to generate panel data for testing
  # Rho is the lag effect
  # if bf = 0 then it is random effect
    # Otherwise fixed effect
GenPanel<-function(N,TT, rho, bf){
  
  id<- c(t(kronecker(matrix(1,1,TT),1:N)))
  
  # Independent variables
  fe <- rnorm(N,0,1) # fixed effect
  fe2<-c(t(kronecker(matrix(1,1,TT+1),fe)))
  x1 <- (bf*fe2 + rnorm((TT+1)*N,0,1))/2
  x2 <- rnorm((TT+1)*N,0,1)
  
  Fit <- x1*0.5 -x2 + fe2 # XB
  Fit <-matrix(Fit, nrow = (TT+1), ncol = N)
  A <-diag(TT+1) 
  A[2:(TT+1),1:TT] <- A[2:(TT+1),1:TT]-rho*diag(TT)
  Mat <- inv(A)
  SIG<- diag(TT+1)
  VAR<- Mat %*% SIG %*% t(Mat) # Variance
  
  Means <- Mat %*% Fit # Means
  
  y <- matrix(, nrow = N, ncol = TT+1) # Dependent variable
  for(i in 1:N){
    y[i,] <- rmvnorm(n=1, mean=t(Means[,i]), sigma=VAR)
  }
  # Some data transformation
  y <- matrix(t(y[,2:(TT+1)]), nrow = TT*N, ncol = 1)
  x1 <- matrix(x1, nrow = (TT+1), ncol = N)
  x1 <- matrix(x1[2:(TT+1),], nrow = TT*N, ncol = 1)
  x2 <- matrix(x2, nrow = (TT+1), ncol = N)
  x2 <- matrix(x2[2:(TT+1),], nrow = TT*N, ncol = 1)
  
  data<-data.frame(id, y,x1, x2)
  return(data)
}


