# Simple linear model

LL_reg <- function(B, y, X) {
  N<-nrow(X)
  b <- B[1:(length(B)-1)]
  lsig <- B[length(B)]
  
  y_std <- (y - X %*% b)/exp(lsig/2)
  # LL enters with a minus
  LL <- sum(0.5*(lsig + log(2*pi) + y_std^2))
  return(LL)
}

RegExample <- function(formula, data) {
  
  X <- model.matrix(formula, data)
  y<-data[,as.character(formula[[2]])]
  
  k <- ncol(X) # no. of coefficients
  res<-optim(rep(0.01,k+1), LL_reg,method = "BFGS", hessian=T,y=y,X=X)
  coefnames <- c(colnames(X), "Sigma")

  b<-res$par
  VC<-solve(res$hessian)
  se<-sqrt(diag(VC))
  
  b[coefnames == "Sigma"] <- exp(b[coefnames == "Sigma"])
  se[coefnames == "Sigma"] <- b[coefnames == "Sigma"]*se[coefnames == "Sigma"]
  
  z<-b/se
  pval<-2*(1-pnorm(abs(z)))
  results<-cbind(b,se,z,pval)
  colnames(results)<-c("b","se","z","p")
  rownames(results)<- coefnames
  print(results,digits=3)
  return(res)
}

# Function to generate panel data for testing
# Rho is the lag effect
# if bf = 0 then it is random effect
# Otherwise fixed effect
GenPanel<-function(N,TT, rho, bf){
  
  id<- c(t(kronecker(matrix(1,1,TT),1:N)))
  
  # Independent variables
  fe <- rnorm(N,0,1) # fixed effect
  fe2<-c(t(kronecker(matrix(1,1,TT+1),fe)))
  x1 <- (bf*fe2 + rnorm((TT+1)*N,0,1))/2
  x2 <- rnorm((TT+1)*N,0,1)
  
  Fit <- x1*0.5 -x2 + fe2 # XB
  Fit <-matrix(Fit, nrow = (TT+1), ncol = N)
  A <-diag(TT+1) 
  A[2:(TT+1),1:TT] <- A[2:(TT+1),1:TT]-rho*diag(TT)
  Mat <- inv(A)
  SIG<- diag(TT+1)
  VAR<- Mat %*% SIG %*% t(Mat) # Variance
  
  Means <- Mat %*% Fit # Means
  
  y <- matrix(, nrow = N, ncol = TT+1) # Dependent variable
  for(i in 1:N){
    y[i,] <- rmvnorm(n=1, mean=t(Means[,i]), sigma=VAR)
  }
  # Some data transformation
  y <- matrix(t(y[,2:(TT+1)]), nrow = TT*N, ncol = 1)
  x1 <- matrix(x1, nrow = (TT+1), ncol = N)
  x1 <- matrix(x1[2:(TT+1),], nrow = TT*N, ncol = 1)
  x2 <- matrix(x2, nrow = (TT+1), ncol = N)
  x2 <- matrix(x2[2:(TT+1),], nrow = TT*N, ncol = 1)
  
  data<-data.frame(id, y,x1, x2)
  return(data)
}